/*
** Module   : WidgetFactoryAdapter.java
** Abstract : provide empty implementations for WidgetFactory interface
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140521 Created initial version
** 002 CA  20140711 Added MESSAGE-AREA widget support.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added support for IMAGE widget creation.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Added support for scroll bar widget creation.
** 006 MAG 20141009 Deleted createMessageLine.
** 007 EVL 20141118 Adding scroll container creation method.
** 008 CA  20150323 Pass the frame ID when creating a label (to properly resolve the font in GUI).
** 009 HC  20150323 Removed business logic from config classes.
** 010 EVL 20150428 Adding drop-down container creation.
*/

package com.goldencode.p2j.ui.client.driver;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Provide default implementations for {@link WidgetFactory} interface.  Any API not overridden
 * by sub-classes will throw a {@link RuntimeException} by default.
 */
public abstract class WidgetFactoryAdapter<O extends OutputManager<?>>
implements WidgetFactory<O>
{
   /** Mapping between abstract and UI-specific widget classes. */
   @SuppressWarnings("rawtypes")
   protected static Map<Class<? extends Widget>, Class<? extends Widget>> classMap = 
      new HashMap<>();
   
   /**
    * A cache of widget constructors having as single parameter a {@link WidgetConfig} instance.
    */
   private Map<Class<? extends Widget<?>>, Constructor<?>> widgetCtorsCache = new HashMap<>();
      
   /**
    * Instantiate UI-specific version of the specified widget class.
    * 
    * @param   <T>
    *          Widget abstract class type.
    * @param   clazz
    *          Widget abstract class for which UI-specific implementation should
    *          be instantiated.
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Widget configuration container.
    *          
    * @return  new UI-specific instance of the requested class. 
    */
   @Override
   public <T extends Widget<O>> T create(Class<T> clazz, WidgetId id, WidgetConfig cfg)
   {
      @SuppressWarnings("unchecked")
      Class<T> implClazz = (Class<T>) classMap.get(clazz);
      
      if (implClazz == null) {
         throw new RuntimeException("Attempt to instantiate class" + clazz + 
                                    " which has no proper mapping");
      }
      
      try
      {
         @SuppressWarnings("unchecked")
         Constructor<T> constructor = (Constructor<T>) widgetCtorsCache.get(implClazz);
         if (constructor == null)
         {
            // find a compatible constructor - two arguments, first assignable to 
            // WidgetId and second assignable to WidgetConfig
            Constructor<?>[] ctors = implClazz.getConstructors();
            
            for (Constructor<?> ctor : ctors)
            {
               Class<?>[] params = ctor.getParameterTypes();
               if (params.length != 2)
               {
                  continue;
               }
               
               if (!WidgetId.class.isAssignableFrom(params[0]))
               {
                  continue;
               }
               
               if (!WidgetConfig.class.isAssignableFrom(params[1]))
               {
                  continue;
               }
               
               if (constructor != null)
               {
                  throw new RuntimeException("Multiple compatible constructors found.");
               }
               
               @SuppressWarnings("unchecked")
               Constructor<T> foundCtor = (Constructor<T>) ctor;
               constructor = foundCtor;
            }
            
            if (constructor == null)
            {
               throw new RuntimeException("No compatible constructor found.");
            }
            
            widgetCtorsCache.put(implClazz, constructor);
         }
         
         // this is an expensive operation in a heavy-duty method, but at least we are not 
         // resolving the c'tor each time a new widget is created.
         return constructor.newInstance(id, cfg);
      }
      catch (Exception e)
      {
         throw new RuntimeException("Attempt to instantiate class" + implClazz + 
         " for class " + clazz + " failed", e);
      }
   }

   /**
    * Create new {@link Label} instance for given text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   text
    *          Label text. 
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    * 
    * @return  new {@link Label} instance.
    */
   @Override
   public Label<O> createLabel(WidgetId id, String text, int frameId)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create new button for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          Button text.
    *          
    * @return  new {@link Button} instance.
    */   
   @Override
   public Button<O> createButton(WidgetId id, Container<O> container, String text)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create new toggle-box for given container and text.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    * @param   text
    *          ToggleBox label text.
    *          
    * @return  new {@link ToggleBox} instance.
    */
   @Override
   public ToggleBox<O> createToggleBox(WidgetId id, Container<O> container, String text)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create new rectangle for given container.
    * 
    * @param   id
    *          The widget's ID.
    * @param   container
    *          Container to which new instance will belong.
    *          
    * @return  new {@link Rect} instance.
    */
   @Override
   public Rect<O> createRectangle(WidgetId id, Container<O> container)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link DropDown} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    *          
    * @return  instance of UI-specific implementation of {@link DropDown}.
    */
   @Override
   public DropDown<O> createDropDown(ComboBox<O> parent)
   {
      return new DropDown<O>(parent);
   }

   /**
    * Create UI-specific {@link ScrollableList} implementation.
    * 
    * @param   parent
    *          Parent {@link ComboBox} widget.
    * @param   model
    *          Scrollable list model.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollableList}.
    */   
   @Override
   public ScrollableList<O> createScrollableList(ComboBox<O> parent, ListModel<String> model)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link ScrollContainer} implementation.
    * 
    * @param   isScrollable
    *          The boolean flag that indicates whether the ScrollContainer has
    *          the scrolling capability.
    * @param   frame 
    *          Frame widget which will own the container.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollContainer<O> createScrollContainer(boolean isScrollable, Frame<O> frame)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link ScrollPane} implementation.
    * 
    * @param   id
    *          The widget's ID.
    * @param   widget
    *          Widget which will be scrolled.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollPane}.
    */
   @Override
   public ScrollPane<O> createScrollPane(WidgetId id, Widget<O> widget)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create UI-specific {@link Label} instance as a copy of given {@link Label}.
    * 
    * @param   id
    *          The widget's ID.
    * @param   labelInstance
    *          Source {@link Label} instance.
    * @param   frameId
    *          The ID of the frame to which the label belongs, or -1 if unknown at the time of the 
    *          creation.
    *          
    * @return  instance of UI-specific implementation of {@link Label}.
    */
   @Override
   public Label<O> createLabel(WidgetId id, Label<O> labelInstance, int frameId)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link RadioButton} instance.
    * 
    * @param   config
    *          Parent config. 
    * @param   label
    *          Radio button label. 
    * @param   value
    *          Radio button data type. 
    * 
    * @return  instance of UI-specific implementation of {@link RadioButton}.
    */   
   @Override
   public RadioButton<O> createRadioButton(RadioSetConfig config,
                                           String         label,
                                           BaseDataType   value)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link Window} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Window}.
    */   
   @Override
   public Window<O> createWindow(WidgetId id)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link StatusLine} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link StatusLine}.
    */   
   @Override
   public StatusLine<O> createStatusLine(WidgetId id)
   {
      throw new RuntimeException("Not implemented.");
   }

   /**
    * Create UI-specific {@link MessageArea} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   lines
    *          Number of visible lines in message area.         
    * 
    * @return  instance of UI-specific implementation of {@link MessageArea}.
    */
   @Override
   public MessageArea<O> createMessageArea(WidgetId id, int lines)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create UI-specific {@link Image} instance.
    * 
    * @param   id
    *          The widget's ID.
    * 
    * @return  instance of UI-specific implementation of {@link Image}.
    */
   @Override
   public Image<O> createImage(WidgetId id)
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Create UI-specific {@link ScrollBar} instance.
    * 
    * @param   id
    *          The widget's ID.
    * @param   orientation
    *          Scroll bar orientation horizontal or vertical.
    *          
    * @return  instance of UI-specific implementation of {@link ScrollBar}.
    */
   public ScrollBar<O> createScrollBar(WidgetId id, ScrollBar.Orientation orientation)
   {
      throw new RuntimeException("Not implemented.");
   }   
}
